@@ -12,18 +12,24 @@ import errno
 import optparse
 import os
 import popen2
-import re
 import shutil
 import signal
 import sys
 import tempfile
 import time
 
-# hghave reserved exit code to skip test
+# reserved exit code to skip test (used by hghave)
 SKIPPED_STATUS = 80
+SKIPPED_PREFIX = 'skipped: '
 
 required_tools = ["python", "diff", "grep", "unzip", "gunzip", "bunzip2", "sed"]
 
+defaults = {
+    'jobs': ('HGTEST_JOBS', 1),
+    'timeout': ('HGTEST_TIMEOUT', 180),
+    'port': ('HGTEST_PORT', 20059),
+}
+
 parser = optparse.OptionParser("%prog [options] [tests]")
 parser.add_option("-C", "--annotate", action="store_true",
     help="output files annotated with coverage")
@@ -36,27 +42,38 @@ parser.add_option("-f", "--first", action="store_true",
 parser.add_option("-i", "--interactive", action="store_true",
     help="prompt to accept changed output")
 parser.add_option("-j", "--jobs", type="int",
-    help="number of jobs to run in parallel")
+    help="number of jobs to run in parallel"
+         " (default: $%s or %d)" % defaults['jobs'])
+parser.add_option("--keep-tmpdir", action="store_true",
+    help="keep temporary directory after running tests"
+         " (best used with --tmpdir)")
 parser.add_option("-R", "--restart", action="store_true",
     help="restart at last error")
 parser.add_option("-p", "--port", type="int",
-    help="port on which servers should listen")
+    help="port on which servers should listen"
+         " (default: $%s or %d)" % defaults['port'])
 parser.add_option("-r", "--retest", action="store_true",
     help="retest failed tests")
 parser.add_option("-s", "--cover_stdlib", action="store_true",
     help="print a test coverage report inc. standard libraries")
 parser.add_option("-t", "--timeout", type="int",
-    help="kill errant tests after TIMEOUT seconds")
+    help="kill errant tests after TIMEOUT seconds"
+         " (default: $%s or %d)" % defaults['timeout'])
 parser.add_option("--tmpdir", type="string",
     help="run tests in the given temporary directory")
 parser.add_option("-v", "--verbose", action="store_true",
     help="output verbose messages")
+parser.add_option("-n", "--nodiff", action="store_true",
+    help="skip showing test changes")
 parser.add_option("--with-hg", type="string",
     help="test existing install at given location")
 
-parser.set_defaults(jobs=1, port=20059, timeout=180)
+for option, default in defaults.items():
+    defaults[option] = int(os.environ.get(*default))
+parser.set_defaults(**defaults)
 (options, args) = parser.parse_args()
 verbose = options.verbose
+nodiff = options.nodiff
 coverage = options.cover or options.cover_stdlib or options.annotate
 python = sys.executable
 
@@ -64,8 +81,15 @@ if options.jobs < 1:
     print >> sys.stderr, 'ERROR: -j/--jobs must be positive'
     sys.exit(1)
 if options.interactive and options.jobs > 1:
-    print >> sys.stderr, 'ERROR: cannot mix -interactive and --jobs > 1'
-    sys.exit(1)
+    print '(--interactive overrides --jobs)'
+    options.jobs = 1
+
+def rename(src, dst):
+    """Like os.rename(), trade atomicity and opened files friendliness
+    for existing destination support.
+    """
+    shutil.copy(src, dst)
+    os.remove(src)
 
 def vlog(*msg):
     if verbose:
@@ -92,10 +116,10 @@ def extract_missing_features(lines):
     '''Extract missing/unknown features log lines as a list'''
     missing = []
     for line in lines:
-        if not line.startswith('hghave: '):
+        if not line.startswith(SKIPPED_PREFIX):
             continue
         line = line.splitlines()[0]
-        missing.append(line[8:])
+        missing.append(line[len(SKIPPED_PREFIX):])
 
     return missing
 
@@ -125,9 +149,10 @@ def check_required_tools():
             print "WARNING: Did not find prerequisite tool: "+p
 
 def cleanup_exit():
-    if verbose:
-        print "# Cleaning up HGTMP", HGTMP
-    shutil.rmtree(HGTMP, True)
+    if not options.keep_tmpdir:
+        if verbose:
+            print "# Cleaning up HGTMP", HGTMP
+        shutil.rmtree(HGTMP, True)
 
 def use_correct_python():
     # some tests run python interpreter. they must use same
@@ -154,7 +179,7 @@ def install_hg():
     # Run installer in hg root
     os.chdir(os.path.join(os.path.dirname(sys.argv[0]), '..'))
     cmd = ('%s setup.py clean --all'
-           ' install --force --home="%s" --install-lib="%s"'
+           ' install --force --prefix="%s" --install-lib="%s"'
            ' --install-scripts="%s" >%s 2>&1'
            % (sys.executable, INST, PYTHONDIR, BINDIR, installerrs))
     vlog("# Running", cmd)
@@ -180,6 +205,20 @@ def install_hg():
     os.environ["PYTHONPATH"] = pythonpath
 
     use_correct_python()
+    global hgpkg
+    hgpkg = _hgpath()
+
+    vlog("# Installing dummy diffstat")
+    f = open(os.path.join(BINDIR, 'diffstat'), 'w')
+    f.write('#!' + sys.executable + '\n'
+            'import sys\n'
+            'files = 0\n'
+            'for line in sys.stdin:\n'
+            '    if line.startswith("diff "):\n'
+            '        files += 1\n'
+            'sys.stdout.write("files patched: %d\\n" % files)\n')
+    f.close()
+    os.chmod(os.path.join(BINDIR, 'diffstat'), 0700)
 
     if coverage:
         vlog("# Installing coverage wrapper")
@@ -257,7 +296,7 @@ def run(cmd):
                        % options.timeout)
     return ret, splitnewlines(output)
 
-def run_one(test, skips):
+def run_one(test, skips, fails):
     '''tristate output:
     None -> skipped
     True -> passed
@@ -266,18 +305,27 @@ def run_one(test, skips):
     def skip(msg):
         if not verbose:
             skips.append((test, msg))
-            sys.stdout.write('s')
-            sys.stdout.flush()
         else:
             print "\nSkipping %s: %s" % (test, msg)
         return None
 
+    def fail(msg):
+        fails.append((test, msg))
+        if not nodiff:
+            print "\nERROR: %s %s" % (test, msg)
+        return None
+
     vlog("# Test", test)
 
     # create a fresh hgrc
     hgrc = file(HGRCPATH, 'w+')
     hgrc.write('[ui]\n')
     hgrc.write('slash = True\n')
+    hgrc.write('[defaults]\n')
+    hgrc.write('backout = -d "0 0"\n')
+    hgrc.write('commit = -d "0 0"\n')
+    hgrc.write('debugrawcommit = -d "0 0"\n')
+    hgrc.write('tag = -d "0 0"\n')
     hgrc.close()
 
     err = os.path.join(TESTDIR, test+".err")
@@ -314,7 +362,9 @@ def run_one(test, skips):
         if os.name == 'nt':
             return skip("shell script")
         # do not try to run non-executable programs
-        if not os.access(testpath, os.X_OK):
+        if not os.path.exists(testpath):
+            return fail("does not exist")
+        elif not os.access(testpath, os.X_OK):
             return skip("not executable")
         cmd = '"%s"' % testpath
 
@@ -328,8 +378,9 @@ def run_one(test, skips):
     if options.timeout > 0:
         signal.alarm(0)
 
+    mark = '.'
+
     skipped = (ret == SKIPPED_STATUS)
-    diffret = 0
     # If reference output file exists, check test output against it
     if os.path.exists(ref):
         f = open(ref, "r")
@@ -337,22 +388,27 @@ def run_one(test, skips):
         f.close()
     else:
         ref_out = []
-    if not skipped and out != ref_out:
-        diffret = 1
-        print "\nERROR: %s output changed" % (test)
-        show_diff(ref_out, out)
     if skipped:
+        mark = 's'
         missing = extract_missing_features(out)
         if not missing:
             missing = ['irrelevant']
         skip(missing[-1])
+    elif out != ref_out:
+        mark = '!'
+        if ret:
+            fail("output changed and returned error code %d" % ret)
+        else:
+            fail("output changed")
+        if not nodiff:
+            show_diff(ref_out, out)
+        ret = 1
     elif ret:
-        print "\nERROR: %s failed with error code %d" % (test, ret)
-    elif diffret:
-        ret = diffret
+        mark = '!'
+        fail("returned error code %d" % ret)
 
     if not verbose:
-        sys.stdout.write('.')
+        sys.stdout.write(mark)
         sys.stdout.flush()
 
     if ret != 0 and not skipped:
@@ -387,7 +443,8 @@ def run_one(test, skips):
         pass
 
     os.chdir(TESTDIR)
-    shutil.rmtree(tmpd, True)
+    if not options.keep_tmpdir:
+        shutil.rmtree(tmpd, True)
     if skipped:
         return None
     return ret == 0
@@ -401,6 +458,8 @@ if not options.child:
 # the tests produce repeatable output.
 os.environ['LANG'] = os.environ['LC_ALL'] = 'C'
 os.environ['TZ'] = 'GMT'
+os.environ["EMAIL"] = "Foo Bar <foo.bar@example.com>"
+os.environ['CDPATH'] = ''
 
 TESTDIR = os.environ["TESTDIR"] = os.getcwd()
 HGTMP = os.environ['HGTMP'] = tempfile.mkdtemp('', 'hgtests.', options.tmpdir)
@@ -408,9 +467,7 @@ DAEMON_PIDS = None
 HGRCPATH = None
 
 os.environ["HGEDITOR"] = sys.executable + ' -c "import sys; sys.exit(0)"'
-os.environ["HGMERGE"]  = ('python "%s" -L my -L other'
-                          % os.path.join(TESTDIR, os.path.pardir,
-                                         'contrib', 'simplemerge'))
+os.environ["HGMERGE"] = "internal:merge"
 os.environ["HGUSER"]   = "test"
 os.environ["HGENCODING"] = "ascii"
 os.environ["HGENCODINGMODE"] = "strict"
@@ -426,9 +483,21 @@ BINDIR = os.path.join(INST, "bin")
 PYTHONDIR = os.path.join(INST, "lib", "python")
 COVERAGE_FILE = os.path.join(TESTDIR, ".coverage")
 
+def _hgpath():
+    cmd = '%s -c "import mercurial; print mercurial.__path__[0]"'
+    hgpath = os.popen(cmd % python)
+    path = hgpath.read().strip()
+    hgpath.close()
+    return path
+
+expecthg = os.path.join(HGTMP, 'install', 'lib', 'python', 'mercurial')
+hgpkg = None
+
 def run_children(tests):
     if not options.with_hg:
         install_hg()
+        if hgpkg != expecthg:
+            print '# Testing unexpected mercurial: %s' % hgpkg
 
     optcopy = dict(options.__dict__)
     optcopy['jobs'] = 1
@@ -461,13 +530,17 @@ def run_children(tests):
     failures = 0
     tested, skipped, failed = 0, 0, 0
     skips = []
+    fails = []
     while fps:
         pid, status = os.wait()
         fp = fps.pop(pid)
         l = fp.read().splitlines()
         test, skip, fail = map(int, l[:3])
-        for s in l[3:]:
+        split = -fail or len(l)
+        for s in l[3:split]:
             skips.append(s.split(" ", 1))
+        for s in l[split:]:
+            fails.append(s.split(" ", 1))
         tested += test
         skipped += skip
         failed += fail
@@ -476,6 +549,11 @@ def run_children(tests):
     print
     for s in skips:
         print "Skipped %s: %s" % (s[0], s[1])
+    for s in fails:
+        print "Failed %s: %s" % (s[0], s[1])
+
+    if hgpkg != expecthg:
+        print '# Tested unexpected mercurial: %s' % hgpkg
     print "# Ran %d tests, %d skipped, %d failed." % (
         tested, skipped, failed)
     sys.exit(failures != 0)
@@ -489,6 +567,9 @@ def run_tests(tests):
         if not options.with_hg:
             install_hg()
 
+            if hgpkg != expecthg:
+                print '# Testing unexpected mercurial: %s' % hgpkg
+
         if options.timeout > 0:
             try:
                 signal.signal(signal.SIGALRM, alarmed)
@@ -513,11 +594,12 @@ def run_tests(tests):
                 tests = orig
 
         skips = []
+        fails = []
         for test in tests:
             if options.retest and not os.path.exists(test + ".err"):
                 skipped += 1
                 continue
-            ret = run_one(test, skips)
+            ret = run_one(test, skips, fails)
             if ret is None:
                 skipped += 1
             elif not ret:
@@ -525,8 +607,9 @@ def run_tests(tests):
                     print "Accept this change? [n] ",
                     answer = sys.stdin.readline().strip()
                     if answer.lower() in "y yes".split():
-                        os.rename(test + ".err", test + ".out")
+                        rename(test + ".err", test + ".out")
                         tested += 1
+                        fails.pop()
                         continue
                 failed += 1
                 if options.first:
@@ -537,12 +620,18 @@ def run_tests(tests):
             fp = os.fdopen(options.child, 'w')
             fp.write('%d\n%d\n%d\n' % (tested, skipped, failed))
             for s in skips:
-                fp.write("%s %s\n" % s) 
+                fp.write("%s %s\n" % s)
+            for s in fails:
+                fp.write("%s %s\n" % s)
             fp.close()
         else:
             print
             for s in skips:
                 print "Skipped %s: %s" % s
+            for s in fails:
+                print "Failed %s: %s" % s
+            if hgpkg != expecthg:
+                print '# Tested unexpected mercurial: %s' % hgpkg
             print "# Ran %d tests, %d skipped, %d failed." % (
                 tested, skipped, failed)
 
